# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------
import unittest
from Xhpc.relocate import *
import pkg_resources


class TestGetNodesPpn(unittest.TestCase):

    def setUp(self) -> None:
        self.args = {
            'job_id': 'X',
            'include': None,
            'exclude': None,
            'localscratch': None,
            'userscratch': False,
            'scratch': False,
            'torque': False,
            'scratching': [],
            'move_from': set(),
            'move_to': set(),
            'mkdir': set(),
            'clear': [],
            'move': True,
            'quiet': False
        }
        self.test_folder = pkg_resources.resource_filename("Xhpc", "test")
        self.folder_of_not_existing = self.test_folder
        self.not_existing = ['%s/n1' % self.test_folder,
                             '%s/n2.txt' % self.test_folder]
        self.folders = ['%s/d1' % self.test_folder,
                        '%s/d2' % self.test_folder]
        for folder in self.folders:
            os.makedirs(folder)
        self.files = ['%s/f1.txt' % self.test_folder,
                      '%s/f2.txt' % self.test_folder]
        for file in self.files:
            with open(file, 'w'):
                pass
        self.set = set(self.not_existing)
        self.set.update(set(self.folders))
        self.set.update(set(self.files))

    def test_get_scratch_path(self):

        self.args['userscratch'] = '/some/path'
        with self.assertRaises(IOError) as cm:
            get_scratch_path(self.args)
        self.assertEqual(
            str(cm.exception),
            'Must specify `--scratch`, `--userscratch` or '
            '`--localscratch to use `--move`')

        self.args['scratch'] = '/some/path'
        with self.assertRaises(IOError) as cm:
            get_scratch_path(self.args)
        self.assertEqual(
            str(cm.exception),
            'Must specify `--scratch`, `--userscratch` or '
            '`--localscratch to use `--move`')

        self.args['localscratch'] = '/localscratch:10GB'
        with self.assertRaises(IOError) as cm:
            get_scratch_path(self.args)
        self.assertEqual(
            str(cm.exception),
            'Must specify `--scratch`, `--userscratch` or '
            '`--localscratch to use `--move`')

    def test_get_scratch_path(self):
        self.assertIsNone(get_scratch_path(self.args))

        self.args['move'] = True
        self.args['userscratch'] = '/some/path'
        obs = get_scratch_path(self.args)
        self.assertEqual('/some/path', obs)

        self.args['scratch'] = '/some/path'
        obs = get_scratch_path(self.args)
        self.assertEqual('/some/path', obs)

        self.args['localscratch'] = '/localscratch:10GB'
        obs = get_scratch_path(self.args)
        self.assertEqual('/localscratch:10GB', obs)

    def test_get_scratching_commands_local(self):
        self.args['localscratch'] = '/localscratch:10GB'
        get_scratching_commands(self.args)
        exp = ['# Define and create a scratch directory',
               'SCRATCH_FOLDER="/localscratch:10GB/${X}"',
               'mkdir -p ${SCRATCH_FOLDER}',
               'cd ${SCRATCH_FOLDER}',
               'echo Working directory is ${SCRATCH_FOLDER}']
        self.assertEqual(exp, self.args['scratching'])

    def test_get_scratching_commands_user(self):
        self.args['userscratch'] = '/some/path'
        get_scratching_commands(self.args)
        exp = ['# Define and create a scratch directory',
               'SCRATCH_FOLDER="/some/path/${X}"',
               'mkdir -p ${SCRATCH_FOLDER}',
               'cd ${SCRATCH_FOLDER}',
               'echo Working directory is ${SCRATCH_FOLDER}']
        self.assertEqual(exp, self.args['scratching'])

    def test_get_scratching_commands(self):
        self.args['scratch'] = '/some/path'
        get_scratching_commands(self.args)
        exp = ['# Define and create a scratch directory',
               'SCRATCH_FOLDER="/some/path/${X}"',
               'mkdir -p ${SCRATCH_FOLDER}',
               'cd ${SCRATCH_FOLDER}',
               'echo Working directory is ${SCRATCH_FOLDER}']
        self.assertEqual(exp, self.args['scratching'])

    def test_get_in_out(self):
        obs = get_in_out(self.set)
        exp = {'files': set(self.files),
               'folders': set(self.folders) | {self.folder_of_not_existing},
               'out': set()}
        # get_in_out(self.set)
        self.assertEqual(exp, obs)

    def test_get_min_files(self):
        files = {
            '/a/b/c/d/e.txt',
            '/a/b/c/d.txt',
            '/a/b/c.txt',
            '/1/2.txt',
            '/1.txt'
        }
        min_folders = {'/a/b/c', '/1'}
        obs = get_min_files(files, min_folders)
        exp = {'/a/b/c.txt', '/1.txt'}
        self.assertEqual(exp, obs)

    def test_get_min_folders(self):
        folders = {
            '/a/b/c/d/e',
            '/a/b/c/d',
            '/a/b/c',
            '/1/2/3/4/5',
            '/1'
        }
        obs = get_min_folders(folders, set())
        exp = {'/a/b/c', '/1'}
        self.assertEqual(exp, obs)

    def test_get_include_commands(self):
        exp = set()
        obs = get_include_commands(self.args)
        self.assertEqual(exp, obs)

        self.args['include'] = ('./a', './1/2/3',)
        obs = get_include_commands(self.args)
        self.assertEqual(exp, obs)

        os.makedirs('./a')
        path = os.path.abspath('./a')
        self.args['include'] = ('./a', './1/2/3',)
        obs = get_include_commands(self.args)
        self.assertEqual({path}, obs)
        path_dir = dirname(path)
        exp = {'mkdir -p ${SCRATCH_FOLDER}%s' % path_dir}
        self.assertEqual(exp, self.args['mkdir'])
        exp = {'rsync -aqru %s/ ${SCRATCH_FOLDER}%s' % (path, path)}
        self.assertEqual(exp, self.args['move_to'])
        exp = {'rsync -aqru ${SCRATCH_FOLDER}%s/ %s' % (path, path)}
        self.assertEqual(exp, self.args['move_from'])
        os.rmdir('./a')

    def test_go_to_work(self):
        go_to_work(self.args)
        exp = ['# Move to the working directory and say it',
               'cd $SLURM_SUBMIT_DIR',
               'echo Working directory is $SLURM_SUBMIT_DIR']
        self.assertEqual(exp, self.args['move_to'])

        self.args['torque'] = True
        go_to_work(self.args)
        exp = ['# Move to the working directory and say it',
               'cd $PBS_O_WORKDIR',
               'echo Working directory is $PBS_O_WORKDIR']
        self.assertEqual(exp, self.args['move_to'])

    def tearDown(self) -> None:
        self.folders = ['%s/d1' % self.test_folder,
                        '%s/d2' % self.test_folder]
        for folder in self.folders:
            os.rmdir(folder)
        self.files = ['%s/f1.txt' % self.test_folder,
                      '%s/f2.txt' % self.test_folder]
        for file in self.files:
            os.remove(file)


if __name__ == '__main__':
    unittest.main()



# ----------------------------------------------------------------------------
# Copyright (c) 2020, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import unittest
import subprocess
from os.path import dirname
from Xhpc.env import *
import pkg_resources
test_folder = pkg_resources.resource_filename("Xhpc", "test")


class TestConda(unittest.TestCase):

    def setUp(self):
        self.env_name = 'testforXhpc'
        self.conda_bin = subprocess.getoutput('which conda')
        cmd = ['conda', 'create', '-n', self.env_name,
               'python', '--no-default-packages', '--yes']
        stdout, stderr = subprocess.Popen(
            cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE).communicate()
        for i in stdout.decode().split('\n'):
            if 'environment location' in i:
                self.env_path = i.strip().split()[-1]
                libs = glob.glob('%s/bin/*' % self.env_path)
                self.conda_exe = set([basename(x) for x in libs])
                break
        self.envs_path = dirname(self.env_path)
        self.conda_bin_path = self.envs_path.replace('envs', 'bin/conda')

        self.dummy_paths = ['%s/path1' % test_folder, '%s/path2' % test_folder]
        self.dummy_executable_paths = []
        self.dummy_executables = {'a', 'b', 'c'}
        for dummy_path in self.dummy_paths:
            os.makedirs(dummy_path)
            for dummy_executable in self.dummy_executables:
                dummy_executable_path = dummy_path + '/' + dummy_executable
                self.dummy_executable_paths.append(dummy_executable_path)
                with open(dummy_executable_path, "w"):
                    pass

    def test_get_conda_path(self):
        obs = get_conda_path(self.env_path)
        self.assertEqual(self.env_path, obs)

        os.environ = {'CONDA_ENVS_PATH': self.envs_path}
        obs = get_conda_path(self.env_name)
        self.assertEqual(self.env_path, obs)

        os.environ = {'CONDA_EXE': self.conda_bin_path}
        obs = get_conda_path(self.env_name)
        self.assertEqual(self.env_path, obs)

        os.environ = {}
        with self.assertRaises(OSError) as cm:
            get_conda_path(self.env_name)
        self.assertEqual(str(cm.exception), 'Make sure conda is installed')

    def test_get_conda_executables(self):
        obs = get_conda_executables('/this/path/surely/does/not/exist')
        self.assertEqual(set(), obs)

        obs = get_conda_executables(self.env_path)
        self.assertEqual(self.conda_exe, obs)

        os.environ = {'CONDA_ENVS_PATH': self.envs_path}
        obs = get_conda_executables(self.env_name)
        self.assertEqual(self.conda_exe, obs)

        os.environ = {'CONDA_EXE': self.conda_bin_path}
        obs = get_conda_executables(self.env_name)
        self.assertEqual(self.conda_exe, obs)

        os.environ = {}
        with self.assertRaises(OSError) as cm:
            get_conda_executables(self.env_name)
        self.assertEqual(str(cm.exception), 'Make sure conda is installed')

    def test_get_path_executables(self):
        os.environ = {'PATH': ':'.join(list(self.dummy_paths))}
        obs = get_path_executables()
        self.assertEqual(self.dummy_executables, obs)

        os.environ = {'PATH': self.dummy_paths[0]}
        obs = get_path_executables()
        self.assertEqual(self.dummy_executables, obs)

        os.environ = {'OTHER_ENV': 'whatever'}
        obs = get_path_executables()
        self.assertEqual(set(), obs)

    def test_get_executables(self):
        os.environ = {'OTHER_ENV': 'whatever'}
        args = {'env': '/this/path/surely/does/not/exist'}
        get_executables(args)
        self.assertEqual(set(), args['executables'])

        args = {'env': self.env_path}
        get_executables(args)
        self.assertEqual(self.conda_exe, args['executables'])

        args = {'env': self.env_name}
        os.environ = {'CONDA_ENVS_PATH': self.envs_path}
        get_executables(args)
        self.assertEqual(self.conda_exe, args['executables'])

        args = {'env': self.env_name}
        os.environ = {'CONDA_EXE': self.conda_bin_path}
        get_executables(args)
        self.assertEqual(self.conda_exe, args['executables'])

        args = {'env': self.env_name}
        os.environ = {}
        with self.assertRaises(OSError) as cm:
            get_executables(args)
        self.assertEqual(str(cm.exception), 'Make sure conda is installed')

    def tearDown(self):
        cmd = [self.conda_bin, 'remove', '-n', self.env_name, '--all', '--yes']
        subprocess.call(cmd)
        for dummy_executable_path in self.dummy_executable_paths:
            os.remove(dummy_executable_path)
        for dummy_path in self.dummy_paths:
            os.rmdir(dummy_path)


if __name__ == '__main__':
    unittest.main()



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import os
import glob
from os.path import basename


def get_conda_path(env: str) -> str:
    """List the executables by parsing the binaries in the bin folder of the
    conda environment. Sometimes, the envs are installed somewhere as set
    out in some env variables. Hence, this tool follows the order of
    precedence of the conda: first will look into CONDA_ENVS_PATH, then will
    look relative to CONDA_EXE and finally, will find using a subprocess. A
    subprocess can take a few seconds to run, so I recommend the user set a
    CONDA_ENVS_PATH variable (e.g., in the `.bashrc` file as explained here:
    https://docs.conda.io/projects/conda/en/latest/user-guide/configuration
    /use-condarc.html#specify-environment-directories-envs-dirs).

    Parameters
    ----------
    env : str
        Conda environment to run the job

    Raises
    ------
    OSError
        If the given env is not an absolute path, or that no useful env
        variable exists for conda that can be used to fond the env binaries.

    Returns
    -------
    path : str
        Path to the set of executables from the conda environment
    """
    path = None
    if env:
        if env[0] == '/':
            path = env
        elif 'CONDA_ENVS_PATH' in os.environ:
            path = '%s/%s' % (os.environ['CONDA_ENVS_PATH'], env)
        elif 'CONDA_EXE' in os.environ:
            path = os.environ['CONDA_EXE'].replace('bin/conda', 'envs/%s' % env)
        else:
            raise OSError('Make sure conda is installed')
    return path


def get_conda_executables(args: dict) -> set:
    """Get the basename of all executables present in the bin folder of the
    conda environment that the job will run in, in order for these
    executables not be affected by scratch/localscratch transfers.

    Parameters
    ----------
    args : dict
        All arguments
        env : str
            Conda environment to run the job

    Returns
    -------
    conda_executables : set
        Set of executables
    """
    conda_executables = set()
    if args['env']:
        args['env'] = get_conda_path(args['env'])
        conda_executables.update(
            [basename(x) for x in glob.glob('%s/bin/*' % args['env'])])
    return conda_executables


def get_path_executables() -> set:
    """Collect the set of files that are located in the paths listed in every
    environment variable containing "PATH". That should include "PATH",
    "PYTHONPATH", and other paths set specifically and that may contains
    files for tools that are ok if moved around (e.g. "GTDBTK_DATA_PATH").
    Indeed, this collected set is used later to avoid changing moving
    binaries from and to the (local)scratch locations.

    Returns
    -------
    path_executables : set
        Path to the set of executables from the user's environment
    """
    path_executables = set()
    for env_variable, env_variable_value in os.environ.items():
        if 'PATH' in env_variable:
            for path in env_variable_value.split(':'):
                path_executables.update(
                    [basename(x) for x in glob.glob('%s/*' % path)])
    return path_executables


def get_executables(args: dict) -> None:
    """Get the basename of all executables present in the bin folder of the
    conda environment that the job will run in and of every environment
    variable containing "PATH". That should include "PATH", "PYTHONPATH",
    and other paths set specifically and that may contains files for tools
    that are ok if moved around (e.g. "GTDBTK_DATA_PATH").
    Indeed, this collected set is used later to avoid changing moving
    binaries from and to the (local)scratch locations, in order for these
    executables not be affected by scratch/localscratch transfers.
    This results in extending the `args` dictionary with the "executables"
    key, pointing to the set of conda and global paths executables.

    Parameters
    ----------
    args : dict
        All arguments
    """
    executables = get_path_executables()
    executables.update(get_conda_executables(args))
    args['executables'] = executables



# ----------------------------------------------------------------------------
# Copyright (c) 2020, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------
import unittest
from unittest.mock import patch
from Xhpc.scratches import *
import pkg_resources
test_folder = pkg_resources.resource_filename("Xhpc", "test")


class TestScratches(unittest.TestCase):

    def setUp(self):
        self.args = {
            'scratch': False, 'userscratch': False, 'config_scratch': False}
        self.path = '%s/out.txt' % test_folder
        self.dir = '%s/config' % test_folder
        os.makedirs(self.dir)
        self.scratch_txt = '%s/scratch.txt' % self.dir
        self.userscratch_txt = '%s/userscratch.txt' % self.dir
        self.content = '/any/thing'
        self.userwork = '${USERWORK}'
        self.scratch_fp = './scratch.txt'
        with open(self.scratch_fp, 'w') as o:
            o.write('/some/scratch/folder\n')
        self.scratch_default = '/cluster/work/jobs'
        self.userscratch_default = '/cluster/work/users/$USER'

    def test_edit_scratch(self):
        obs = edit_scratch(self.args, self.scratch_fp, 'scratch')
        self.assertEqual('/some/scratch/folder', obs)

    @patch('builtins.input', return_value='y')
    def test_edit_scratch_y(self, input):
        self.args['config_scratch'] = True
        obs = edit_scratch(self.args, self.scratch_fp, 'scratch')
        self.assertEqual(self.scratch_default, obs)
        obs = edit_scratch(self.args, self.scratch_fp, 'userscratch')
        self.assertEqual(self.userscratch_default, obs)

    @patch('builtins.input', return_value='y')
    def test_get_scratches(self, input):
        self.assertIsNone(get_scratches(self.args, self.dir))

        self.args['userscratch'] = True
        get_scratches(self.args, self.dir)
        self.assertEqual(self.userscratch_default, self.args['userscratch'])
        self.assertTrue(isfile(self.userscratch_txt))

        self.args['scratch'] = True
        get_scratches(self.args, self.dir)
        self.assertEqual(self.scratch_default, self.args['scratch'])
        self.assertTrue(isfile(self.scratch_txt))

    def test_create_scratch(self):
        os.environ = {'USERWORK': 1}
        obs = create_scratch(self.args, self.scratch_fp, 'userscratch')
        self.assertEqual('${USERWORK}', obs)
        self.args['config_scratch'] = True
        self.assertEqual('${USERWORK}', obs)

    @patch('builtins.input', return_value='y')
    def test_create_scratch_input(self, input):
        os.environ = {'USERWORK': 1}
        obs = create_scratch(self.args, self.scratch_fp, 'scratch')
        self.assertEqual('/cluster/work/jobs', obs)
        os.environ = {}
        obs = create_scratch(self.args, self.scratch_fp, 'userscratch')
        self.assertEqual('/cluster/work/users/$USER', obs)
        obs = create_scratch(self.args, self.scratch_fp, 'scratch')
        self.assertEqual('/cluster/work/jobs', obs)

    @patch('builtins.input', return_value='y')
    def test_get_scratch_area_y(self, input):
        obs = get_scratch_area("scratch", 'a')
        self.assertEqual('a', obs)
        obs = get_scratch_area("userscratch", 'b')
        self.assertEqual('b', obs)

    @patch('builtins.input', return_value='n')
    def test_get_scratch_area_n_is_not(self, input):
        with self.assertRaises(IOError) as cm:
            get_scratch_area("scratch", '/not/a/valid/path')
        self.assertEqual(str(cm.exception),
                         'Enter a valid "scratch" folder path...')

        with self.assertRaises(IOError) as cm:
            get_scratch_area("userscratch", '/not/a/valid/path')
        self.assertEqual(str(cm.exception),
                         'Enter a valid "userscratch" folder path...')

    @patch('builtins.input', return_value='./scratch.txt')
    def test_get_scratch_area_n_is(self, input):
        obs = get_scratch_area("scratch", self.scratch_fp)
        self.assertEqual(self.scratch_fp, obs)

    @patch('builtins.input', return_value='y')
    def test_get_scratch_y(self, input):
        obs = get_scratch("scratch")
        self.assertEqual('/cluster/work/jobs', obs)
        obs = get_scratch("userscratch")
        self.assertEqual('/cluster/work/users/$USER', obs)

    @patch('builtins.input', return_value='n')
    def test_get_scratch_n_is_not(self, input):
        with self.assertRaises(IOError) as cm:
            get_scratch("scratch")
        self.assertEqual(str(cm.exception),
                         'Enter a valid "scratch" folder path...')

        with self.assertRaises(IOError) as cm:
            get_scratch("userscratch")
        self.assertEqual(str(cm.exception),
                         'Enter a valid "userscratch" folder path...')

    @patch('builtins.input', return_value='./scratch.txt')
    def test_get_scratch_n_is(self, input):
        obs = get_scratch("scratch")
        self.assertEqual(self.scratch_fp, obs)
        obs = get_scratch("localscratch")
        self.assertEqual(self.scratch_fp, obs)

    @patch('builtins.print')
    def test_write_scratches_print(self, prints):
        write_scratches(self.path, self.userwork)
        lines = []
        with open(self.path) as f:
            for line in f:
                lines.append(line.strip())
        self.assertEqual([self.userwork], lines)
        print_out = prints.call_args_list[0][0][0]
        self.assertEqual(print_out, 'Written: %s' % self.path)
        os.remove(self.path)

    def test_write_scratches(self):
        write_scratches(self.path, self.content)
        lines = []
        with open(self.path) as f:
            for line in f:
                lines.append(line.strip())
        self.assertEqual([self.content], lines)
        os.remove(self.path)

    def tearDown(self) -> None:
        os.remove(self.scratch_fp)
        if isfile(self.scratch_txt):
            os.remove(self.scratch_txt)
        if isfile(self.userscratch_txt):
            os.remove(self.userscratch_txt)
        if os.path.isdir(self.dir):
            os.rmdir(self.dir)


if __name__ == '__main__':
    unittest.main()







# https://askubuntu.com/questions/430194/no-such-file-or-directory-when-executing-usr-bin-time
# http://www.linuxintheshell.com/2013/02/26/phobia-2016-turkce-altyazili-hd-izle-filmi-izle-full-izle/

# too assess memory:
# Empirtical Approach to run the tool with more and more data
# https://researchcomputing.princeton.edu/support/knowledge-base/memory
#   - collect all reports
#   - parse the Memorz Utilized...
# e.g. using seff

# Or dynamically get the currently running JOBID:
#   - by running series of squeue -U franckl
#   - by writing for this assessment work all the JOBID/<hostname> in a file
# -> then ssh <hostname> and htop -u $USER


# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------
import os
import sys
import glob
import subprocess
import pandas as pd
from datetime import datetime
from os.path import abspath, isfile, isdir


def init_args(args: dict) -> None:
    """Initialize key/value pairs in the args dict to be filled is necessary.

    Parameters
    ----------
    args : dict
        All arguments
    """
    args['scratching'] = []
    args['mkdir'] = set()
    args['move_to'] = set()
    args['move_from'] = set()
    args['clear'] = []


def get_sinfo_pd(args: dict, sinfo_dir: str) -> None:
    """Collect the nodes info from the output of sinfo if the user asks
    for Xhpc to allocate specific nodes ot the job being created,
    or if the user wishes to update sinfo (it is run once per day by default
    but can be called again if one needs to use nodes that may become
    available in the course of the day).
    The `args` dictionary will be extended with a "sinfo_pd" key, pointing
    to a pd.DataFrame containing the påer-node sinfo output for an
    extensive set of parameters.

    Parameters
    ----------
    args : dict
        All arguments. Here only the following keys are of interest:
            sinfo: bool
                Whether to update node usage info (overwrites `~/.sinfo` too)
            allocate : bool
                Get current machine usage to allocate suitable nodes/memory
            torque : bool
                Switch from Slurm to Torque
    sinfo_dir : str
        Path to the .sinfo folder where is the most up-to-date sinfo file
    """
    sinfo_pd = pd.DataFrame()
    # # -------------------------------------------------------------
    # # -----------               DELETE                  -----------
    # # -------------------------------------------------------------
    # fp = '/Users/franck/programs/Xhpc/Xhpc/test/snap.txt'
    # sinfo_pd = pd.read_csv(fp, sep='\t')
    # args['sinfo_pd'] = sinfo_pd
    # return sinfo_pd
    # # -------------------------------------------------------------
    # # -------------------------------------------------------------
    if args['torque']:
        raise IOError('No node allocation yet avail for PBS/Torque')
    # if the user wants to update sinfo or needs to Xhpc to
    # allocate specific nodes/cores for the job.
    elif args['sinfo'] or args['allocate']:
        # then we need to rely on the latest sinfo data, which is located in
        # the dot folder '.sinfo' and in the file time stamped for today

        # -------------------------------------------------------------
        # -----------               CHANGE                  -----------
        # -------------------------------------------------------------
        # fp = '/Users/franck/programs/Xhpc/Xhpc/test/snap.txt'
        fp = '%s/%s.tsv' % (sinfo_dir, str(datetime.now().date()))
        # -------------------------------------------------------------
        # -------------------------------------------------------------

        # if there is no file today or if the user wants to update sinfo
        if args['sinfo'] or not isfile(fp):
            # create the dot folder '.sinfo' if needed
            if not isdir(sinfo_dir):
                os.makedirs(sinfo_dir)
            # collect the sinfo data
            sinfo_pd = collect_sinfo_pd()
            # remove previous files for former days
            remove_previous_fps(sinfo_dir)
            # write it out as the latest, today's file
            sinfo_pd.to_csv(fp, index=False, sep='\t')
        # if there is a file already for today
        else:
            # just read it in!
            sinfo_pd = pd.read_csv(fp, sep='\t')
    args['sinfo_pd'] = sinfo_pd


def decode_sinfo_stdout(sinfo_stdout: bytes) -> list:
    """Transform the raw bytes text sinfo stdout output into rows represented
    as  lists (of lists).

    Parameters
    ----------
    sinfo_stdout : bytes
        per node sinfo subprocess stdout for an extensive set of parameters

    Returns
    -------
    sinfo_list : list of lists
        Per-node sinfo output for an extensive set of parameters
    """
    sinfo_decoded = sinfo_stdout.decode()
    sinfo_split = sinfo_decoded.strip().split('\n')
    sinfo_list = [row.split() for row in sinfo_split]
    return sinfo_list


def collect_sinfo_pd() -> pd.DataFrame:
    """Run sinfo as a subprocess and collect the sinfo subprocess results as
    a pandas table.

    Returns
    -------
    sinfo_pd : pd.DataFrame
        per node sinfo subprocess stdout for an extensive set of  parameters
    """
    sinfo_stdout = run_sinfo_subprocess()
    sinfo_out = decode_sinfo_stdout(sinfo_stdout)
    sinfo_pd = pd.DataFrame(sinfo_out, columns=[
        'node', 'partition', 'status', 'cpu_load', 'cpus',
        'socket', 'cores', 'threads', 'mem', 'free_mem'])
    return sinfo_pd


def run_sinfo_subprocess() -> bytes:
    """Run sinfo as a subprocess.

    Returns
    -------
    sinfo_stdout : bytes
        per node sinfo subprocess stdout for an extensive set of  parameters
    """
    fmt = 'NodeList:10,'
    fmt += 'Partition:10,'
    fmt += 'StateLong:10,'
    fmt += 'CPUsLoad:10,'
    fmt += 'CPUsState:12,'
    fmt += 'Sockets:4,'
    fmt += 'Cores:4,'
    fmt += 'Threads:4,'
    fmt += 'Memory:12,'
    fmt += 'FreeMem:12'
    cmd = ['sinfo', '--Node', '-h', '-O', fmt]
    # run sinfo and get the stdout
    sinfo_stdout, stderr = subprocess.Popen(
        ' '.join(cmd),
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        shell=True
    ).communicate()
    if stderr:
        raise FileNotFoundError('Using Slurm? `sinfo` command not found')
    return sinfo_stdout


def remove_previous_fps(sinfo_dir: str) -> None:
    """Clear the .sinfo folder from every already-existsing / previous sinfo
    file in order to only keep the most up-to-date sinfo data (written after
    this function is called).

    Parameters
    ----------
    sinfo_dir : str
        Path to the .sinfo folder where is the most up-to-date sinfo file
    """
    for previous_fp in glob.glob('%s/*.tsv' % sinfo_dir):
        os.remove(previous_fp)


def get_job_fp(args: dict) -> None:
    """Get the path to the file that will be written to contains all the
    slurm (or torque) job directives and the job commands it self.
    This results in extending the `args` dictionary with the "job_fp"
    key, pointing to the path to the job file. By default, this file is the
    input extended with date and time info and `.slm` for slurm (or `.pbs`
    for torque).

    Parameters
    ----------
    args : dict
        All arguments. Here only the following keys are of interest:
            job: str
                Job name
            output_fp: str
                Output script path
            torque: bool
                Adapt to Torque
    """
    # if the user did not specify a path to a script output file using option
    # `--o-script`, create one based on the input and the date
    if not args['output_fp']:
        # format the date and time of now
        now_time = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
        # add `.slm` (or `.pbs`) to a concatenation of path, job name and time
        job_rad = '%s/%s_%s' % (abspath('.'), args['job'], now_time)
        if args['torque']:
            args['job_fp'] = '%s.pbs' % job_rad
        else:
            args['job_fp'] = '%s.slm' % job_rad
    # if the output filename was provided, just use it
    else:
        args['job_fp'] = args['output_fp']


def get_output_dir(args: dict) -> None:
    """Get the absolute path of the output directory.
    This results in extending the `args` dictionary with the "output_dir"
    key, pointing to the output directory (full path) to use for the job.

    Parameters
    ----------
    args : dict
        All arguments. Here only the following keys are of interest:
        out_dir: str
            Output directory
    """
    out_dir = args['out_dir']
    if isdir(out_dir):
        args['output_dir'] = abspath(out_dir)
    else:
        args['output_dir'] = abspath('.')


def get_first_line(path: str) -> str:
    """Get the first line of a file.

    Parameters
    ----------
    path : str
        Path to a file

    Returns
    -------
    ret: str
        First line of the file
    """
    ret = ""
    with open(path) as f:
        for line in f:
            ret = line.strip()
            break
    return ret


def get_tmpdir(args: dict) -> bool:
    """Define the path to a temporary folder depending on the user options
    such as the use of scratch or localscratch, or on the presence of an
    already existing TMPDIR folder, or on whether the user specified a tmp
    folder to use. It creates the folder and export an environment variable
    TMPDIR for it.

    Parameters
    ----------
    args : dict
        All arguments. Here only the following keys are of interest:
            tmp: str
                Alternative temp folder
            workdir: str
                Working directory
            move : bool
                Move files/folders to chosen scratch location
            torque: bool
                Adapt to Torque
    """
    # define the temporary folder
    if args['tmp']:
        tmpdir = args['tmp'].rstrip('/')
    elif args['move']:
        # if this is set, then there will be a SCRATCH_FOLDER folder created
        tmpdir = '${SCRATCH_FOLDER}/tmpdir'
    elif 'TMPDIR' in os.environ:
        tmpdir = '${TMPDIR}'
    else:
        return None  # if no TMPDIR is set, keep the default machine behavior

    tmpdir += '_' + args['job']
    if args['torque']:
        tmpdir += '_${PBS_JOBID}'
    else:
        tmpdir += '_${SLURM_JOB_ID}'

    # set command to create the temporary folder
    args['tmp'] = [
        '\n# create and export the temporary directory',
        'mkdir -p %s' % tmpdir,
        'export TMPDIR="%s"' % tmpdir,
        'echo Temporary directory is ${TMPDIR}']
    args['clear'].append('rm -rf ${TMPDIR}')


def check_content(args: dict) -> None:
    """Ask user to check for the job script content, which is composed of
    "directives", "preamble", and "commands"

    Parameters
    ----------
    args : dict
        All arguments. Here only the following keys are of interest:
            verif: bool
                Print script to stdout and query user for sanity
            directives : list
                Commands composing the job's directives
            preamble : list
                Commands composing the job's preamble
            commands : list
                Commands composing the actual job
    """
    if args['verif']:
        for part in ['directives', 'preamble', 'scratching', 'tmp', 'mkdir',
                     'move_to', 'commands', 'move_from', 'clear']:
            if part not in args or not args[part]:
                continue
            print('------------------------%s' % ('-' * len(part)))
            print("Please check the job's %s:" % part)
            print('------------------------%s' % ('-' * len(part)))
            for command in args[part]:
                print(command)
            ret = input('\n\nContinue to write?: <[y]/n>\n')
            if ret == 'n':
                print('Exiting\n')
                sys.exit(1)


def get_lines(part) -> list:
    """

    Parameters
    ----------
    part : list or set
        Commands to write out for a section

    Returns
    -------
    lines : list
        Commands list to write out for a section (if was a set, a sorted list)
    """
    if isinstance(part, list):
        lines = part
    elif isinstance(part, set):
        lines = sorted(part)
    else:
        raise TypeError('Commands are collected either as list or set')
    return lines


def write_out(args: dict) -> None:
    """Write the actual .pbs / slurm .sh script based on
    the info collected from the command line.

    Parameters
    ----------
    args : dict
        All arguments, including:
            directives : list
                Commands composing the job's directives
            preamble : list
                Commands composing the job's preamble
            commands : list
                Commands composing the actual job
            stat: bool
                Whether to prepend `/usr/bin/time -v` to every script command
            quiet : bool
                Do not print anything
    """
    with open(args['job_fp'], 'w') as o:
        for part in ['directives', 'preamble', 'scratching', 'tmp', 'mkdir',
                     'move_to', 'commands', 'move_from', 'clear']:
            if part not in args or not args[part]:
                continue
            next_line = False
            lines = get_lines(args[part])
            for line_ in lines:
                line = line_
                if args['stat'] and part == 'commands':
                    if next_line or not line_.strip():
                        line = line_
                    else:
                        line = '/usr/bin/time -v %s' % line_
                o.write('%s\n' % line)
                next_line = False
                if line.endswith('\\'):
                    next_line = True
            if part in args:
                o.write('# ------ %s END ------\n' % part)
            o.write('\n')
        o.write('echo "Done!"\n')
        if not args['quiet']:
            print('Written:', args['job_fp'])


def show_config(args: dict, config_fp: str) -> None:
    """

    Parameters
    ----------
    args : dict
        All arguments. Here only the following key is of interest:
        scratch: str
            Path to scratch folder (to move files and compute)
        userscratch: str
            Path to user scratch folder (to move files and compute)
        email_address : str
            email address of the user.
    config_fp : str
        Path to the file that may (or not) contains an email address
    """
    if isfile(config_fp):
        print('* email address config file:',
              subprocess.getoutput('cat %s' % config_fp))
    else:
        print('Email address config not yet created')
        sys.exit(1)

    for scratch in ['scratch', 'userscratch']:
        if args[scratch]:
            scratch_fp = config_fp.replace('config.txt', '%s.txt' % scratch)
            if isfile(scratch_fp):
                print('* %s folder config file:' % scratch,
                      subprocess.getoutput('cat %s' % scratch_fp))
            else:
                print('%s config not created (run Xhpc using `--%s`)' % (
                    scratch, scratch))
                sys.exit(1)


def sys_exit(args: dict):
    """Quit if some arguments are not passed by user.
    Note that these could be set using required=True in the click options,
    but this is not suited to letting this tools reach the show_config().

    Parameters
    ----------
    args : dict
        All arguments, including:
            input_fp : str
                Input script path (or double-quoted command)
            job: str
                Job name
    """
    if not args['input_fp']:
        print('Option "-i" (or "--i-script") is mandatory')
        sys.exit(0)
    if not args['job']:
        print('Option "-j" (or "--i-job") is mandatory')
        sys.exit(0)



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import os
import subprocess
from os.path import expanduser
import pkg_resources as pkg_res

from Xhpc.email import get_email_address
from Xhpc.scratches import get_scratches
from Xhpc.env import get_executables
from Xhpc.directives import get_directives
from Xhpc.preamble import get_preamble
from Xhpc.cmd import get_commands
from Xhpc.relocate import get_relocation
from Xhpc.io_utils import (
    init_args, get_job_fp, get_output_dir, get_sinfo_pd, get_tmpdir,
    write_out, check_content, show_config, sys_exit)


def xhpc(**args) -> None:
    """
    Main script to go from .sh script to the Slurm or Torque output script file.
    Writes a Slurm/Torque script with directives and possibly re-localization
    of file if working on a (local)scratch folder.

    Parameters
    ----------
    args : dict
        All arguments, including:
            input_fp : str
                Input script path (or double-quoted command)
            job: str
                Job name
            output_fp: str
                Output script path
            account : str
                Name of the account
            partition: str
                Partition name
            out_dir: str
                Output directory
            env: str
                Conda environment to run the job
            nnodes: int
                Number of nodes
            cpus: int
                Number of CPUs
            time: str
                Wall time limit
            tmp: str
                Alternative temp folder
            mem: tuple
                Requested memory
            mem_per_cpu: tuple
                Requested memory
            nodes: tuple
                Node names
            workdir: str
                Working directory
            scratch: str
                Path to scratch folder (to move files and compute)
            userscratch: str
                Path to user scratch folder (to move files and compute)
            localscratch: tuple
                Use localscratch with the provided memory amount (in gb)
            clear_scratch : bool
                Whether to cleat the scratch area at the end of the job or not
            include : tuple
                Folder to not move to and from scratch using rsync (must exist)
            exclude : tuple
                Relative path(s) within input folder(s) to not move in scratch
            stdout : bool
                Rename stdout (and stderr) with job name and ID
            email: bool
                Send email (always if fail)
            run: bool
                Run the job output script
            move : bool
                Move files/folders to chosen scratch location
            abspath : bool
                Change the existing paths in command line to absolute paths
            verif: bool
                Print script to stdout and query user for sanity
            stat: bool
                Whether to prepend `/usr/bin/time -v` to every script command
            gpu: bool
                Query a gpu (experimental)
            torque: bool
                Adapt to Torque
            config_email: bool
                Show the current email and potentially edit it
            config_scratch : bool
                Show current scratches folder and/or edit it
            sinfo: bool
                Print sinfo in stdout (and update `~/.sinfo`)
            allocate : bool
                Get current machine usage to allocate suitable nodes/memory
            show_config : bool
                Show current configurations (email and scratches)
            quiet : bool
                Do not print anything
    """
    init_args(args)
    config_dir = '%s/user/%s' % (pkg_res.resource_filename("Xhpc", ""),
                                 os.environ['USER'])
    config_fp = '%s/config.txt' % config_dir
    if args['show_config']:
        show_config(args, config_fp)
    else:
        sys_exit(args)
        get_email_address(args, config_fp)  # get email address
        get_scratches(args, config_dir)  # get paths to scratch folders
        get_sinfo_pd(args, '%s/.sinfo' % expanduser('~'))  # get node usage
        get_job_fp(args)  # get path of output job file
        get_output_dir(args)  # get absolute path of the output directory
        get_executables(args)  # get set of executables in current environment
        get_directives(args)  # prepare job directives
        get_commands(args)  # get job command lines
        get_preamble(args)  # set environment and working directory
        get_tmpdir(args)  # set temporary directory
        get_relocation(args)   # arrange file movement
        # write the psb file to provide to "qsub"
        check_content(args)  # print-based, visual checks
        write_out(args)
        if args['run']:
            if not args['quiet']:
                print('Launched command: /bin/sh %s' % args['job_fp'])
            subprocess.call(['qsub', args['job_fp']])



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------
import os
from os.path import isfile
from Xhpc.io_utils import get_first_line


def edit_scratch(args: dict, scratch_fp: str, scratch: str) -> str:
    """
    Checks that the content of the existing scratches file are valid paths.

    Parameters
    ----------
    args : dict
        All arguments. Here only the following key is of interest:
            config_scratch : bool
                Show current scratches folder and/or edit it
    scratch_fp : str
        Path to the file containing the scratch folder
    scratch : str
        "scratch" or "userscratch"

    Returns
    -------
    scratch_folder : str
        Scratch folder
    """
    # parse the first line of the config file
    scratch_folder = get_first_line(scratch_fp)
    if args['config_scratch']:
        print("Registered %s: %s" % (scratch, scratch_folder))
        scratch_folder = create_scratch(args, scratch_fp, scratch)
    return scratch_folder


def get_scratches(args: dict, config_dir: str) -> None:
    """
    Collect the scratch folder paths from the scratch config file.

    Parameters
    ----------
    args : dict
        All arguments. Here only the following key is of interest:
            scratch: str
                Path to scratch folder (to move files and compute)
            userscratch: str
                Path to user scratch folder (to move files and compute)
            config_scratch : bool
                Show current scratches folder and/or edit it
    config_dir : str
       Path to the folder that may (or not) contains the scratch paths files
    """
    for scratch in ['scratch', 'userscratch']:
        if args[scratch]:
            scratch_fp = '%s/%s.txt' % (config_dir, scratch)
            if isfile(scratch_fp):
                args[scratch] = edit_scratch(args, scratch_fp, scratch)
            else:
                args[scratch] = create_scratch(args, scratch_fp, scratch)


def create_scratch(args: dict, scratch_fp: str, scratch: str) -> str:
    """Collect the scratch folders interactively from the user
    and write it somewhere it can be reused.

    Parameters
    ----------
    args : dict
        All arguments. Here only the following key is of interest:
            config_scratch : bool
                Show current scratches folder and/or edit it
    scratch_fp : str
        Path to the file containing the scratch folder
    scratch : str
        "scratch" or "userscratch"

    Returns
    -------
    scratch_folder : str
        Scratch folder
    """
    if args['config_scratch']:
        scratch_folder = get_scratch(scratch)
    elif scratch == 'userscratch' and 'USERWORK' in os.environ:
        scratch_folder = '${USERWORK}'
    else:
        scratch_folder = get_scratch(scratch)
    write_scratches(scratch_fp, scratch_folder)
    return scratch_folder


def get_scratch_area(scratch: str, default: str) -> str:
    """Collect the scratch/userscratch folder interactively from the user.

    Parameters
    ----------
    scratch : str
        "scratch" or "userscratch"
    default : str
        Default scratch folder on SAGA

    Returns
    -------
    scratch_folder : str
        Scratch folder
    """
    ret = input('(default [on SAGA]: "%s"): ' % default)
    if ret == 'y':
        scratch_folder = default
    else:
        if not isfile(ret):
            raise IOError('Enter a valid "%s" folder path...' % scratch)
        scratch_folder = ret
    return scratch_folder


def get_scratch(scratch: str) -> str:
    """Collect the scratch folders interactively from the user.

    Parameters
    ----------
    scratch : str
        "scratch" or "userscratch"

    Returns
    -------
    scratch_folder : str
        Scratch folder
    """
    print('Please enter %s location ("y" to confirm default)' % scratch)
    if scratch == 'userscratch':
        scratch_folder = get_scratch_area(scratch, '/cluster/work/users/$USER')
    else:
        scratch_folder = get_scratch_area(scratch, '/cluster/work/jobs')
    return scratch_folder


def write_scratches(scratch_fp: str, scratch_folder: str) -> None:
    """Write the scratch folder somewhere it can be reused.

    Parameters
    ----------
    scratch_fp : str
        Path to the config file
    scratch_folder : str
        Scratch folder
    """
    with open(scratch_fp, 'w') as o:
        o.write('%s\n' % scratch_folder)
    print('Written: %s' % scratch_fp)
    if scratch_folder == '${USERWORK}':
        print('Written automatically since your machine presets $USERNAME')



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import os
from os.path import abspath


def add_env(args: dict) -> None:
    """Add the preamble for the job.

    Parameters
    ----------
    args : dict
        All arguments. Here only the following keys are of interest:
            env : str
                Conda environment to run the job
    """
    args['preamble'] = [
        '# general environment info / behaviour',
        'uname -a',  # Show some machine specs for reproducibility
        'set -e',    # Exit the script on any error
    ]
    # if a conda environment is used
    if args['env']:
        # activate the env
        env = ['\n# active conda environment',
               'echo "Conda environment is %s"' % args['env']]
        if args['saga']:
            env.extend([
                'module load Anaconda3/2019.03',
                'export PS1=\\$',
                'source ${EBROOTANACONDA3}/etc/profile.d/conda.sh',
                'conda deactivate &>/dev/null'])
        env.append('conda activate %s' % args['env'])
        args['preamble'].extend(env)


def add_workdir(args: dict) -> None:
    """Add the job's environment variable setting the working directory.

    Parameters
    ----------
    args : dict
        All arguments. Here only the following keys are of interest:
            workdir: str
                Working directory
            torque: bool
                Adapt to Torque
    """
    if args['workdir']:
        export = "export SLURM_SUBMIT_DIR='%s'" % args['workdir']
        if args['torque']:
            export = "export PBS_O_WORKDIR='%s'" % args['workdir']
        args['preamble'].extend(['\n# change working directory', export])


def add_tmpdir(args: dict) -> bool:
    """Define the path to a temporary folder depending on the user options
    such as the use of scratch or localscratch, or on the presence of an
    already existing TMPDIR folder, or on whether the user specified a tmp
    folder to use. It creates the folder and export an environment variable
    TMPDIR for it.

    Parameters
    ----------
    args : dict
        All arguments. Here only the following keys are of interest:
            tmp: str
                Alternative temp folder
            workdir: str
                Working directory
            move : bool
                Move files/folders to chosen scratch location
            torque: bool
                Adapt to Torque
    Returns
    -------
    tmp : bool
        Whether a temporary folder is set or not
    """
    # define the temporary folder
    if args['tmp']:
        tmpdir = args['tmp'].rstrip('/')
    elif args['move']:
        # if this is set, then there will be a SCRATCH_FOLDER folder created
        tmpdir = '${SCRATCH_FOLDER}/tmpdir'
    elif 'TMPDIR' in os.environ:
        tmpdir = '${TMPDIR}'
    else:
        return False  # if no TMPDIR is set, keep the default machine behavior

    tmpdir += '/' + args['job']
    if args['torque']:
        tmpdir += '_${PBS_JOBID}'
    else:
        tmpdir += '_${SLURM_JOB_ID}'

    # set command to create the temporary folder
    args['preamble'].extend([
        '# create and export the temporary directory',
        'mkdir -p %s' % tmpdir,
        'export TMPDIR="%s"' % tmpdir,
        'echo Temporary directory is ${TMPDIR}'
    ])
    args['clear'].append('rm -rf ${TMPDIR}')
    return True


def add_procs_nodes(args: dict) -> None:
    """Get the number of nodes and processors actually used by the job.

    Parameters
    ----------
    args : dict
        All arguments. Here only the following keys are of interest:
            torque: bool
                Adapt to Torque
    """
    # Calculate the number of processors/nodes allocated to this run.
    if args['torque']:
        args['preamble'].append('NPROCS=`wc -l < ${PBS_NODEFILE}`')
        args['preamble'].append('NNODES=`uniq ${PBS_NODEFILE} | wc -l`')
    else:
        args['preamble'].append('NPROCS=${SLURM_NTASKS}')
        args['preamble'].append('NNODES=${SLURM_NNODES}')
    args['preamble'].append('echo Use ${NPROCS} procs on ${NNODES} nodes')


def add_echoes(args: dict) -> None:
    """Add things to echo in the job preamble that will be available in the
    stdout.

    Parameters
    ----------
    args : dict
        All arguments. Here only the following keys are of interest:
            input_fp : str
                Input script path (or double-quoted command)
            workdir: str
                Working directory
            torque: bool
                Adapt to Torque
    """
    args['preamble'].extend([
        '\n# echo some info about the job',
        'echo Running on host `hostname`',
        'echo Time is `date`',
        'echo Directory is `pwd`'
    ])
    add_procs_nodes(args)
    stdout = 'echo Job stdout is %s' % args['std_path'],
    stderr = 'echo Job stderr is %s' % args['std_path']
    args['preamble'].extend(['%s.out' % stdout, '%s.err' % stderr])
    args['preamble'].append('echo Job script: %s' % abspath(args['input_fp']))


def get_preamble(args: dict) -> None:
    """Get lines to be written as preamble to the job (not the directives):
    - load conda environment for the job
    - setup temporary directory
    - echo important things for the stdout file

    Parameters
    ----------
    args : dict
        All arguments. Here only the following keys are of interest:
            torque: bool
                Adapt to Torque
    """
    args['preamble'] = list()
    add_env(args)  # environment variables
    # set an alternative workdir for the job
    add_workdir(args)
    # set preamble on the job environment
    add_echoes(args)
